/* File MSNBTP.C
 * Bootp requestor
 *
 * Copyright (C) 1991, University of Waterloo.
 * Copyright (C) 1991, Trustees of Columbia University in the
 *  City of New York.  Permission is granted to any individual or
 *  institution to use, copy, or redistribute this software as long as
 *  it is not sold for profit and this copyright notice is retained.
 *
 * Original version created by Erick Engelke of the University of
 *  Waterloo, Waterloo, Ontario, Canada.
 * Adapted and modified for MS-DOS Kermit by Joe R. Doupnik, 
 *  Utah State University, jrd@cc.usu.edu, jrd@usu.Bitnet.
 *
 * Last edit
 * 6 Sept 1991
 *
 *   BOOTP - Boot Protocol (RFC 854)
 *
 *   These extensions get called if bootphost is set to an IP address or
 *   to 0xffffffff.
 *
 */

#include "msntcp.h"
#include "msnlib.h"

/*
 * Bootstrap Protocol (BOOTP)
 * Based on RFC 951.
 */

#define BOOTREQUEST 	1	/* bootp.bp_op */
#define BOOTREPLY	2

/*
 * structure for send and receives
 */
typedef struct bootp {
	byte	 bp_op;		/* packet op code / message type. */
	byte	 bp_htype;	/* hardware address type, 1 = 10 mb ethernet */
	byte	 bp_hlen;	/* hardware address len, eg '6' for 10mb eth */
	byte	 bp_hops;	/* client sets to zero, optionally used by
				   gateways in cross-gateway booting. */
	longword bp_xid;	/* transaction ID, a random number */
	word	 bp_secs;	/* filled in by client, seconds elapsed since
				   client started trying to boot. */
	word	 bp_spare;
	longword bp_ciaddr;	/* client IP address filled in by client */
				/*  if known*/
	longword bp_yiaddr;	/* 'your' (client) IP address
				   filled by server if client doesn't know */
	longword bp_siaddr;	/* server IP address returned in bootreply */
	longword bp_giaddr;	/* gateway IP address,
				   used in optional cross-gateway booting. */
	byte	 bp_chaddr[16];	/* client hardware address, filled by client */
	byte	 bp_sname[64];	/* optional server host name, null terminated*/

	byte	 bp_file[128];	/* boot file name, null terminated string
				   'generic' name or null in bootrequest,
				   fully qualified directory-path
				   name in bootreply. */
	byte	 bp_vend[ 64];	/* optional vendor-specific area */
};


extern struct bootp _bootp;

/*
 * UDP port numbers, server and client.
 */
#define	IPPORT_BOOTPS	67
#define	IPPORT_BOOTPC	68


/******** the following is stolen from NCSA which came from CUTCP *********/
/* I have not implemented these, but someone may wish to in the future so */
/* I kept them around.                                                    */
/**************************************************************************/

/*
 * "vendor" data permitted for Stanford boot clients.
 */
struct vend {
	byte 	 v_magic[4];	/* magic number */
	longword v_flags;	/* flags/opcodes, etc. */
	byte 	 v_unused[56];	/* currently unused */
};

#define	VM_STANFORD	"STAN"	/* v_magic for Stanford */
#define VM_RFC1048      "\x63\x82\x53\x63" /* magic cookie for BOOTP */
					/* high byte listed first */

/* v_flags values */
#define	VF_PCBOOT	1	/* an IBMPC or Mac wants environment info */
#define	VF_HELP		2		/* help me, I'm not registered */
#define TAG_BOOTFILE_SIZE       13     /* tag used by vend fields rfc 1048 */

/* global variables */
longword bootphost = 0xffffffffL;	/* broadcast */
word bootptimeout = 30;
word bootpon = 0;

extern longword set_timeout();


/*
 * dobootpc - Checks global variables bootptimeout, bootphost
 *             if no host specified, the broadcast address
 *             returns 0 on success and sets ip address
 */
int dobootp()
{
    udp_Socket bsock;
    longword sendtimeout, bootptmo;
    word magictimeout;
    word len, templen;
    struct bootp sendbootp;			/* outgoing data */
    struct bootp bootp;				/* incoming data */
    struct bootp register * sbp, * rbp;
    int status;
    longword xid;
    byte *p, *t;
    int dummy;

    if (pktdevclass == PD_SLIP) return (-1);

	sbp = &sendbootp;			/* registered addresses */
	rbp = &bootp;
	memset((byte *)sbp, 0, sizeof(struct bootp));
	memset((byte *)rbp, 0, sizeof(struct bootp));

	outs("\r\n Requesting a Bootp server ");
    /* We must get Waterloo TCP to use IP address 0 for sending */
    xid = my_ip_addr;   /* a unique value coming from the ethernet card */
    my_ip_addr = 0;

    if (udp_open(&bsock, IPPORT_BOOTPC, bootphost, IPPORT_BOOTPS, NULL) == 0)
	{
        outs("\n\r Unable to find a bootp server");
        return (-1);
	}

    bootptmo = set_timeout(bootptimeout);
    magictimeout = (xid & 7) + 7;  		/* between 7 and 14 seconds */

    sbp->bp_op = BOOTREQUEST;
    sbp->bp_htype = pktdevclass;
		    /* Copy into position the Magic Number used by Bootp */
    bcopy(VM_RFC1048, sbp->bp_vend, 4);		/* high byte first */

    if (pktdevclass == PD_ETHER) sbp->bp_hlen = 6;

    sbp->bp_xid = xid;
    sbp->bp_secs = intel16(1);

    bcopy(eth_addr, sbp->bp_chaddr, sizeof(eth_address));

    while (1) {
	sock_fastwrite(&bsock, (byte *)sbp, sizeof(struct bootp));
        sbp->bp_secs += magictimeout;			/* for next time */
        sendtimeout = set_timeout(magictimeout += (xid > 5) & 7);
	outs(".");				/* progress tics */
        while (chk_timeout(sendtimeout) == 0) {
            if (chk_timeout(bootptmo))
	    		{
			outs(" Timed out, sorry");
                	goto give_up;
			}
            sock_tick(&bsock, &status);
            if (len = sock_dataready(&bsock)) {

                /* got a response, lets consider it */
                templen = sock_fastread(&bsock, (byte *)rbp, 
			sizeof(struct bootp));
                if (templen < sizeof(struct bootp)) {
                    /* too small, not a bootp packet */
		    bcopy((byte *)rbp, 0, sizeof(struct bootp));
                    continue;
                }

                /* we must see if this is for us */
		if (rbp->bp_xid != sbp->bp_xid) {
		    bcopy((byte *)rbp, 0, sizeof(struct bootp));
                    continue;
                }

                /* we must have found it */
		my_ip_addr = intel(rbp->bp_yiaddr);

		if (*(long *)rbp->bp_vend == *(long *)sbp->bp_vend)
			{			/* if magic cookies match */
		    /*RFC1048 complient BOOTP vendor field */
		    /* Based heavily on NCSA Telnet BOOTP */
		    p = &rbp->bp_vend[4]; /* Point just after vendor field */

		    while (*p != 255) {
			switch(*p) {
                          case 0: /* Nop Pad character */
                                 p++;
                                 break;
                          case 1: /* Subnet Mask */
				 sin_mask = intel(*(longword *)(&p[2]));
				 /* and fall through */
			  case 2: /* Time offset */
				 p += *(p+1) + 2;
				 break;
			  case 3: /* gateways */
				for (len = 0; len < *(p+1); len += 4)
				  arp_add_gateway(NULL,
				     intel(*(longword*)(&p[2 + len])));
				  		/* and fall through */
			  case 4: /*time servers */
				  		/* fall through */
                          case 5: /* IEN=116 name server */
                                 p +=*(p+1)+2;
                                 break;
			  case 6: /* Domain Name Servers (BIND) */
				for (len = 0; len < *(p+1); len += 4)
				    add_server(&last_nameserver,
					MAX_NAMESERVERS, def_nameservers,
					    intel(*(longword*)(&p[2 + len])));
						/* and fall through */
			  case 7: /* log server */
				 p += *(p+1)+2;
				 break;
			  case 8: /* cookie server */
				 for (len = 0; len < *(p+1); len += 4)
#ifdef KERMIT
	;	/* omit cookie server with Kermit */
#else
				     add_server(&last_cookie, MAX_COOKIES,
					cookie, intel(*(longword*)(&p[2+len])));
#endif /* KERMIT */
				 /* and fall through */
                          case 9: /* lpr server */
                          case 10: /* impress server */
                          case 11: /* rlp server */
				   p +=*(p+1)+2;
                                   break;
			  case 12: /* Client Hostname */
				  bcopy(&p[2], hostname, MAX_STRING);
				  hostname[MAX_STRING - 1] = 0;
				  p += *(p+1)+2;
                                  break;
                          case 255:
                                   break;
                          default:
                                   p +=*(p+1)+2;
                                   break;
                        } /* end of switch */
                     } /* end of while */
                }/* end of RFC_1048 if */
                goto give_up;
            }
        }
    }
give_up:
    sock_close(&bsock);
    if (my_ip_addr != 0) outs(" succeeded.\r\n");
    return (my_ip_addr == 0);  /* return 0 on success */

sock_err:
    /* major network error if UDP fails */
    sock_close(&bsock);
    return (-1);
}
